import requests
from bs4 import BeautifulSoup
from collections import Counter
from urllib.parse import urljoin, urlparse
from typing import Dict, List, Any
import time
import json
from fake_useragent import UserAgent
from .mobile_analyzer import MobileFriendlinessAnalyzer

class SEOAnalyzer:
    def __init__(self):
        self.ua = UserAgent()
        self.mobile_analyzer = MobileFriendlinessAnalyzer()

    def _get_page_content(self, url: str) -> tuple[requests.Response, BeautifulSoup]:
        """Fetch page content and return response and parsed HTML"""
        headers = {'User-Agent': self.ua.random}
        response = requests.get(url, headers=headers)
        response.raise_for_status()
        return response, BeautifulSoup(response.text, 'html.parser')

    def _extract_keywords_from_sources(self, soup: BeautifulSoup) -> List[str]:
        """Extract keywords from meta keywords, JSON-LD schema, Microdata, and RDFa"""
        keywords = set()

        # Extract from meta keywords
        meta_keywords = soup.find('meta', {'name': 'keywords'})
        if meta_keywords and meta_keywords.get('content'):
            keywords.update([k.strip().lower() for k in meta_keywords['content'].split(',')])

        # Extract from JSON-LD
        json_ld_scripts = soup.find_all('script', {'type': 'application/ld+json'})
        for script in json_ld_scripts:
            try:
                data = json.loads(script.string)
                # Recursively extract keywords from JSON-LD data
                def extract_keywords_from_json(json_data):
                    if isinstance(json_data, dict):
                        for key, value in json_data.items():
                            if key.lower() in ['keywords', 'about', 'description', 'name', 'articlesection', 'genre', 'tag', 'topic', 'subject', 'category']:
                                if isinstance(value, list):
                                    keywords.update([k.strip().lower() for k in value if isinstance(k, str)])
                                elif isinstance(value, str):
                                    keywords.update([k.strip().lower() for k in value.split(',')])
                            elif isinstance(value, (dict, list)):
                                extract_keywords_from_json(value)
                    elif isinstance(json_data, list):
                        for item in json_data:
                            extract_keywords_from_json(item)

                # Process the JSON-LD data recursively
                extract_keywords_from_json(data)
            except (json.JSONDecodeError, AttributeError):
                continue

        # Extract from Microdata
        microdata_elements = soup.find_all(attrs={'itemprop': ['keywords', 'about', 'description', 'name']})
        for element in microdata_elements:
            content = element.get('content') or element.string
            if content:
                if isinstance(content, str):
                    keywords.update([k.strip().lower() for k in content.split(',')])

        # Extract from RDFa
        rdfa_elements = soup.find_all(attrs={'property': True})
        for element in rdfa_elements:
            prop = element.get('property', '').lower()
            if any(keyword in prop for keyword in ['keyword', 'about', 'description', 'title']):
                content = element.get('content') or element.string
                if content and isinstance(content, str):
                    keywords.update([k.strip().lower() for k in content.split(',')])

        return list(keywords)

    def _analyze_keyword_density(self, soup: BeautifulSoup) -> Dict[str, Any]:
        """Calculate keyword density and usage statistics from page content"""
        # Extract visible text
        text = ' '.join([s.strip() for s in soup.stripped_strings])
        text_lower = text.lower()
        words = text_lower.split()
        word_count = len(words)

        if word_count == 0:
            return {
                'total_words': 0,
                'keyword_stats': {},
                'general_density': {}
            }

        # Get keywords from various sources
        target_keywords = self._extract_keywords_from_sources(soup)

        # Calculate keyword statistics
        keyword_stats = {}
        for keyword in target_keywords:
            count = text_lower.count(keyword)
            if count > 0:
                # Calculate density as a percentage
                density_percentage = (count / word_count) * 100
                keyword_stats[keyword] = {
                    'count': count,
                    'density_percentage': round(density_percentage, 2),  # Round to 2 decimal places
                    'density': count / word_count,  # Keep original density for backward compatibility
                    'density_score': round(density_percentage * 3, 2),  # New density score at scale 3
                    'locations': [i for i in range(len(text_lower)) if text_lower.startswith(keyword, i)]
                }

        # Calculate general word density (top 10 most common words)
        counts = Counter(words)
        general_density = {}
        for word, count in counts.most_common(10):
            density_percentage = (count / word_count) * 100
            general_density[word] = {
                'count': count,
                'density_percentage': round(density_percentage, 2),  # Round to 2 decimal places
                'density': count / word_count,  # Keep original density for backward compatibility
                'density_score': round(density_percentage * 3, 2)  # New density score at scale 3
            }

        # Calculate combined density score and percentage (using only keyword_stats)
        combined_density_percentage = sum(stat['density_percentage'] for stat in keyword_stats.values()) if keyword_stats else 0
        combined_density_score = combined_density_percentage / 2.5 if combined_density_percentage else 0

        return {
            'total_words': word_count,
            'keyword_stats': keyword_stats,
            'general_density': general_density,
            'combined_density_score': round(combined_density_score, 2),  # Round to 2 decimal places
            'combined_density_percentage': round(combined_density_percentage, 2)  # Round to 2 decimal places
        }

    def _analyze_meta_data(self, soup: BeautifulSoup) -> Dict[str, str]:
        """Extract meta information from the page"""
        meta_data = {
            'title': soup.title.string if soup.title else None,
            'description': None,
            'keywords': None
        }

        for meta in soup.find_all('meta'):
            name = meta.get('name', '').lower()
            if name == 'description':
                meta_data['description'] = meta.get('content')
            elif name == 'keywords':
                meta_data['keywords'] = meta.get('content')

        return meta_data

    def _analyze_headers(self, soup: BeautifulSoup) -> Dict[str, int]:
        """Count header tags (h1-h6)"""
        return {f'h{i}': len(soup.find_all(f'h{i}')) for i in range(1, 7)}

    def _analyze_links(self, soup: BeautifulSoup, base_url: str) -> Dict[str, List[str]]:
        """Analyze internal and external links"""
        base_domain = urlparse(base_url).netloc
        internal_links = []
        external_links = []

        for link in soup.find_all('a'):
            href = link.get('href')
            if not href:
                continue

            full_url = urljoin(base_url, href)
            parsed_url = urlparse(full_url)
            if parsed_url.netloc == base_domain and full_url != base_url:
                internal_links.append(full_url)
            else:
                external_links.append(full_url)

        return {
            'internal_links': internal_links,
            'external_links': external_links
        }

    def _analyze_images(self, soup: BeautifulSoup) -> Dict[str, Any]:
        """Analyze image tags and their alt attributes"""
        images = soup.find_all('img')
        missing_alt_images = []
        for img in images:
            if not img.get('alt'):
                image_info = {
                    'src': img.get('src', 'unknown'),
                    'location': str(img.parent.name) if img.parent else 'unknown'
                }
                missing_alt_images.append(image_info)

        return {
            'total': len(images),
            'missing_alt': len(missing_alt_images),
            'missing_alt_details': missing_alt_images
        }

    def _get_canonical_url(self, soup: BeautifulSoup) -> str:
        """Extract canonical URL if present"""
        canonical = soup.find('link', {'rel': 'canonical'})
        return canonical['href'] if canonical else None

    def _analyze_structured_data(self, soup: BeautifulSoup) -> Dict[str, Any]:
        """Analyze structured data (JSON-LD, Microdata, RDFa) on the page"""
        structured_data = {
            'json_ld': [],
            'microdata': [],
            'rdfa': [],
            'recommendations': []
        }

        # Check JSON-LD
        json_ld_scripts = soup.find_all('script', {'type': 'application/ld+json'})
        for script in json_ld_scripts:
            try:
                data = json.loads(script.string)
                structured_data['json_ld'].append(data)
            except (json.JSONDecodeError, AttributeError):
                structured_data['recommendations'].append(
                    'Invalid JSON-LD script found - please validate the syntax'
                )

        # Check Microdata
        microdata_elements = soup.find_all(itemscope=True)
        for element in microdata_elements:
            item_type = element.get('itemtype')
            if item_type:
                structured_data['microdata'].append({
                    'type': item_type,
                    'properties': [prop.get('itemprop') for prop in element.find_all(attrs={'itemprop': True})]
                })

        # Check RDFa
        rdfa_elements = soup.find_all(attrs={'property': True})
        rdfa_types = set()
        for element in rdfa_elements:
            vocab = element.get('vocab') or element.find_parent(attrs={'vocab': True})
            if vocab:
                rdfa_types.add(vocab)
                structured_data['rdfa'].append({
                    'vocab': vocab,
                    'property': element.get('property')
                })

        # Generate recommendations
        if not structured_data['json_ld'] and not structured_data['microdata'] and not structured_data['rdfa']:
            structured_data['recommendations'].append(
                'No structured data found - consider adding JSON-LD markup for better search engine understanding'
            )
        
        if len(structured_data['json_ld']) > 0:
            structured_data['recommendations'].append(
                f"Found {len(structured_data['json_ld'])} JSON-LD schema(s)"
            )
        
        if len(structured_data['microdata']) > 0:
            structured_data['recommendations'].append(
                f"Found {len(structured_data['microdata'])} Microdata schema(s)"
            )

        if len(structured_data['rdfa']) > 0:
            structured_data['recommendations'].append(
                f"Found {len(structured_data['rdfa'])} RDFa markup(s)"
            )

        return structured_data

    def _generate_recommendations(self, metrics: Dict) -> Dict[str, Any]:
        """Generate detailed SEO recommendations based on analysis"""
        recommendations = {
            'meta': [],
            'headers': [],
            'images': [],
            'structured_data': metrics['structured_data']['recommendations'],
            'mobile': metrics['mobile_friendliness']['recommendations']
        }

        # Title recommendations
        if not metrics['meta_data']['title']:
            recommendations['meta'].append({
                'type': 'error',
                'message': 'Missing title tag',
                'action': 'Add a title tag to the page'
            })
        elif len(metrics['meta_data']['title']) < 30:
            recommendations['meta'].append({
                'type': 'warning',
                'message': f'Title tag is too short ({len(metrics["meta_data"]["title"])} characters)',
                'action': 'Extend title length to 30-60 characters',
                'current_value': metrics['meta_data']['title']
            })
        elif len(metrics['meta_data']['title']) > 60:
            recommendations['meta'].append({
                'type': 'warning',
                'message': f'Title tag is too long ({len(metrics["meta_data"]["title"])} characters)',
                'action': 'Reduce title length to 30-60 characters',
                'current_value': metrics['meta_data']['title']
            })

        # Meta description recommendations
        if not metrics['meta_data']['description']:
            recommendations['meta'].append({
                'type': 'error',
                'message': 'Missing meta description',
                'action': 'Add a meta description'
            })
        elif len(metrics['meta_data']['description']) < 120:
            recommendations['meta'].append({
                'type': 'warning',
                'message': f'Meta description is too short ({len(metrics["meta_data"]["description"])} characters)',
                'action': 'Extend meta description length to 120-160 characters',
                'current_value': metrics['meta_data']['description']
            })
        elif len(metrics['meta_data']['description']) > 160:
            recommendations['meta'].append({
                'type': 'warning',
                'message': f'Meta description is too long ({len(metrics["meta_data"]["description"])} characters)',
                'action': 'Reduce meta description length to 120-160 characters',
                'current_value': metrics['meta_data']['description']
            })

        # Header recommendations
        if metrics['headers']['h1'] == 0:
            recommendations['headers'].append({
                'type': 'error',
                'message': 'Missing H1 heading',
                'action': 'Add an H1 heading to the page'
            })
        elif metrics['headers']['h1'] > 1:
            recommendations['headers'].append({
                'type': 'warning',
                'message': f'Multiple H1 headings found ({metrics["headers"]["h1"]} headings)',
                'action': 'Use only one H1 heading per page'
            })

        # Image recommendations
        if metrics['images']['missing_alt'] > 0:
            recommendations['images'].append({
                'type': 'warning',
                'message': f'{metrics["images"]["missing_alt"]} image(s) missing alt text',
                'action': 'Add descriptive alt text to images',
                'details': {
                    'total_images': metrics['images']['total'],
                    'images_with_alt': metrics['images']['total'] - metrics['images']['missing_alt'],
                    'images_missing_alt': metrics['images']['missing_alt'],
                    'missing_images_details': metrics['images']['missing_alt_details']
                }
            })

        return recommendations

    def analyze_page(self, url: str) -> Dict[str, Any]:
        """Perform complete SEO analysis of a single page"""
        start_time = time.time()
        
        try:
            response, soup = self._get_page_content(url)
            load_time = time.time() - start_time

            metrics = {
                'url': url,
                'meta_data': self._analyze_meta_data(soup),
                'keyword_density': self._analyze_keyword_density(soup),
                'headers': self._analyze_headers(soup),
                **self._analyze_links(soup, url),
                'images': self._analyze_images(soup),
                'page_speed': {
                    'load_time': f"{load_time:.2f}s",
                    'response_time': f"{response.elapsed.total_seconds():.2f}s"
                },
                'canonical': self._get_canonical_url(soup),
                'structured_data': self._analyze_structured_data(soup),
                'mobile_friendliness': self.mobile_analyzer.analyze(soup)
            }

            metrics['seo_recommendations'] = self._generate_recommendations(metrics)
            return metrics

        except Exception as e:
            return {
                'url': url,
                'error': str(e)
            }